#include <xen/xen.h>
#include "xen/x86_32.h"
#include "xen/x86_64.h"

#include "shared.h"
#include "list.h"

/* convert megabytes to pages and back */
#define MB_TO_PG(x) (x << 8)
#define PG_TO_MB(x) (x >> 8)

enum xenmode {
    XENMODE_32   = 32,
    XENMODE_PAE  = 36,
    XENMODE_64   = 64,
};

union xencallbacks {
    struct xen_callback_x86_32        *cb32;
    uint64_t                          *cb64;
};

union xentraps {
    struct trap_info_x86_32           *tr32;
    struct trap_info_x86_64           *tr64;
};

union xenshinfo {
    struct shared_info_x86_32         *sh32;
    struct shared_info_x86_64         *sh64;
    void                              *ptr;
};

union xenrunstate {
    struct vcpu_runstate_info_x86_32  *rs32;
    struct vcpu_runstate_info_x86_64  *rs64;
};

union xenguestcontext {
    struct vcpu_guest_context_x86_32  *ct32;
    struct vcpu_guest_context_x86_64  *ct64;
    void                              *ptr;
};

struct xenvcpu {
    union xenrunstate                 rs;
    union xenguestcontext             ct;
    struct vcpu_time_info             *time;
    uint8_t                           *evtchn_upcall_pending;
    uint8_t                           *evtchn_upcall_mask;
    uint32_t                          ipi_event;
};

struct xenclock {
    uint32_t wc_version;
    uint32_t wc_sec;
    uint32_t wc_nsec;
};

struct xenvm {
    /* kvm state */
    kvm_context_t kvm;
    enum xenmode mode;
    struct kvm_regs regs;
    struct kvm_sregs sregs;
    int regs_up2date;
    int sregs_up2date;
    int regs_dirty;
    int sregs_dirty;

    /* vm core file mapping */
    char       corefile[256];
    int        corefd;
    size_t     coresize;
    void       *coremap;

    /* memory */
    void       *memory;
    uint64_t   pg_total;
    uint64_t   pg_emu;
    uint64_t   pg_m2p;
    uint64_t   pg_guest;
    uint64_t   mfn_emu;
    uint64_t   mfn_m2p;
    uint64_t   mfn_guest;

    /* parameters */
    char       *kernel;
    char       *ramdisk;
    char       *cmdline;

    /* logging */
    int        condev_works;
    int        debug;
    FILE       *logfile;
    int        nostderr;

    /* timekeeping */
    int                   tsc_shift;
    uint32_t              tsc_mul_frac;
    int                   update_time;
    uint64_t              boot;
    uint64_t              now;
    uint64_t              oneshot;
    uint64_t              period;

    /* xen emu state */
    struct elf_binary     *emu;
    uint64_t              emu_va;
    uint64_t              emu_stack;
    struct descriptor_32  *gdt;
    struct tss_32         *tss_32;
    struct tss_64         *tss_64;
    uint32_t              *m2p_32;
    uint64_t              *m2p_64;
    uint64_t              mfn_map;        /* map cache pgtable (32bit) */
    uint64_t              mfn_pgtables;   /* pgtable pool (64bit)      */
    uint32_t              idx_pgtables;

    /* xen shinfo ptrs */
    struct xenclock       *xen_wc;

    /* xen guest state */
    uint64_t              console_pfn;
    uint64_t              xenstore_pfn;
    union xenshinfo       shinfo;
    struct xenvcpu        vcpu[1];
    struct grant_entry    *grant_table;
    uint32_t              virq[NR_VIRQS];

    /* xen devices */
    char                  *disk;
    char                  *cdrom;
    char                  *vfb;
    char                  *network;

    /* event channels */
    int                   evtchnd;
    uint32_t              next_event;
    uint32_t              console_event;
    uint32_t              xenstore_event;

    /* ipc */
    uint32_t              domid;
    char                  name[64];
    char                  vm[48];
    struct xs_handle      *xenstore;
    struct list_head      callbacks[2]; /* index: enum xenner_cb_type */

    /* statistics */
    uint64_t usr_hcalls[XEN_HCALL_MAX];
    uint64_t *emu_hcalls;
    uint64_t *emu_faults;
};

typedef int (*xenner_callback)(struct xenvm *xen, int handle, void *cb_data);

enum xenner_cb_type {
    XENNER_CB_IOFD   = 0,
    XENNER_CB_EVTCHN = 1,
};

struct xenner_callback {
    enum xenner_cb_type type;
    int                 handle;  /* fd, evtchn port, ... */
    char                *name;
    void                *cb_data;
    xenner_callback     callback;
    struct list_head    list;
};

/* xenner.c */
void alive(void);
uint64_t get_systime(void);
void setup_timer(struct xenvm *xen);
void update_time(struct xenvm *xen, int vcpu);
int raise_event(struct xenvm *xen, int vcpu, int port);
void wait_event(struct xenvm *xen, int vcpu);

int xenner_register_cb(struct xenvm *xen, enum xenner_cb_type type,
		       char *name, int handle, void *data,
		       int (*callback)(struct xenvm *xen, int fd, void *data));
int xenner_callback_evtchn(struct xenvm *xen, int port);

void print_guest_stack(struct xenvm *xen, uint64_t rsp);
void print_emu_stack(struct xenvm *xen, uint64_t rsp);
void vm_kill(struct xenvm *vm, char *reason, uint64_t guest_esp);
void xenner_cleanup(struct xenvm *xen);

void need_regs(struct xenvm *xen);
void need_sregs(struct xenvm *xen);
void flush_regs(struct xenvm *xen);
void flush_sregs(struct xenvm *xen);

/* build.c */
int xen_emu_map(struct xenvm *xen);
int domain_builder(struct xenvm *vm);

/* hypercalls.c */
int do_xen_hypercall(struct xenvm *vm, int nr);
void hypercall_stats(struct xenvm *xen);

/* cpuid.c */
void setup_cpuid(struct xenvm *xen);
void set_msr(struct xenvm *xen, int vcpu, uint32_t index, uint64_t data);
uint64_t get_msr(struct xenvm *xen, int vcpu, uint32_t index);
void print_msrs(struct xenvm *xen, int vcpu);

/* x86_emu.c */
int do_instr_emu(struct xenvm *xen);
int do_general_protection(struct xenvm *xen);
int do_page_fault(struct xenvm *xen);

/* debug.c */
int parse_asm(char *filename);
int do_debug_trap(struct xenvm *xen);

/* log.c */
/*
 * Message lewels:
 *  0 - FIXME.  Nuff sayed ;)
 *  1 - informational init stuff (which is rarely called
 *      and doesn't flood the screen when enabled).
 *  2 - default for debug messages.
 *  3 - debug messages causing *excessive* flood.
 */
#define d0printf(fmt, ...) dbgprintf(xen, 0, fmt, __VA_ARGS__)
#define d1printf(fmt, ...) dbgprintf(xen, 1, fmt, __VA_ARGS__)
#define d2printf(fmt, ...) dbgprintf(xen, 2, fmt, __VA_ARGS__)
#define d3printf(fmt, ...) dbgprintf(xen, 3, fmt, __VA_ARGS__)
void loginit(char *name);
void dbgprintf(struct xenvm *xen, int msglevel, char *fmt, ...)
    __attribute__ ((format(printf, 3, 4)));
void conhyp_print(struct xenvm *xen, char *buf, int len);
void condev_print(struct xenvm *xen, char *buf, int len);
void conemu_print(struct xenvm *xen, char *buf);
void banner_print(struct xenvm *xen, char *msg);

/* vmcore.c */
int vmcore_init(struct xenvm *xen);
void vmcore_fini(struct xenvm *xen);
int vmcore_write_headers(struct xenvm *xen);

/* tsctest.c */
void tsc_calibrate(int *shift, uint32_t *mul_frac);

/* console.c */
int console_init(struct xenvm *xen);
int console_fini(struct xenvm *xen);

/* xenstore.c */
int xenstore_init_early(struct xenvm *xen);
int xenstore_init_late(struct xenvm *xen);
int xenstore_shutdown(struct xenvm *xen, char *action);
void xenstore_fini(struct xenvm *xen);

/* evtchn.c */
int evtchn_init(struct xenvm *xen);
int evtchn_port(struct xenvm *xen, char *name);
int evtchn_notify(struct xenvm *xen, int port);
void evtchn_fini(struct xenvm *xen);

/* statistics.c */
void print_fault_stats(FILE *fp, uint64_t *faults, uint64_t *last);
void print_hcall_stats(FILE *fp, uint64_t *hcalls, uint64_t *last);
